"""
Papai-specific functions and data schema.
"""

import io
import os
import warnings
from typing import TypedDict

import pandas as pd
from google.oauth2 import service_account

from papai_unified_storage.storage import Storage, filesystem
from papai_unified_storage.utils import generate_temporary_filename, joinpath


class Parquet(TypedDict):
    bucket_name: str
    """
    workspace bucket name.
    """
    step_name: str
    """
    friendly parquet name.
    """
    object_name: str
    """
    path to the parquet file.
    """


class Bucket(TypedDict):
    bucket_name: str
    """
    workspace bucket name.
    """
    step_name: str
    """
    friendly folder name that we consider being the bucket.
    """
    settings: dict
    """
    store connection settings to external buckets.
    """


class Registry(TypedDict):
    bucket_name: str
    """
    workspace bucket name.
    """
    artefacts_path: str
    """
    path to the artefact folder.
    """
    registry_name: str
    """
    friendly registry name.
    """


def _get_io_config(key: str, match_value: str, list_io: list, io_name: str) -> dict:
    for io_ in list_io:
        if io_[key] == match_value:
            return io_
    raise ValueError(f"Could not find {io_name} with name {match_value}")


def _get_bucket_fs(
    bucket_name: str, papai_fs: Storage, list_buckets: list[Bucket]
) -> tuple[Storage, str]:
    """Get a filesystem to interact with a specific bucket.

    Parameters
    ----------
    bucket_name : str
        Name of the bucket to interact with.
    papai_fs : Storage
        Default filesystem to interact with the bucket.
    list_buckets : list[Bucket]
        List of all bucket configurations.

    Returns
    -------
    filesystem, path_prefix : tuple[Storage, str]
        Filesystem that allows interaction with the bucket.
        Path prefix to be used in papai internal bucket. If you are using an
        external bucket, this will be an empty string.
    """
    bucket = _get_io_config("step_name", bucket_name, list_buckets, "dataset")
    settings = bucket["settings"]

    if "virtual_bucket_path" in settings:
        return papai_fs, settings["virtual_bucket_path"]

    available_protocols = {
        "AZURE_OBJECT_STORAGE_SETTINGS": _get_abfs,
        "S3_OBJECT_STORAGE_SETTINGS": _get_s3fs,
        "GC_OBJECT_STORAGE_SETTINGS": _get_gcsfs,
    }
    protocol = available_protocols[settings["kind"]]

    return filesystem(protocol, settings["storage_options"]), ""


def _get_abfs(settings: dict) -> Storage:
    account_name = settings["account_name"]
    account_key = settings["account_key"]
    return filesystem("abfs", account_name=account_name, account_key=account_key)


def _get_gcsfs(settings: dict) -> Storage:
    credentials_dict = {
        "type": "service_account",
        "private_key_id": settings["private_key_id"],
        "private_key": settings["private_key"],
        "client_email": settings["client_email"],
        "token_uri": "https://oauth2.googleapis.com/token",
    }
    credentials = service_account.Credentials.from_service_account_info(
        credentials_dict
    )
    return filesystem("gcsfs", token=credentials)


def _get_s3fs(settings: dict) -> Storage:
    endpoint = settings["endpoint"]
    access_key = settings["access_key"]
    secret_key = settings["secret_key"]
    return filesystem("s3fs", key=access_key, secret=secret_key, endpoint=endpoint)


def _get_bucket_fs_with_prefix(
    bucket_name: str, papai_fs: Storage, list_buckets: list[Bucket]
) -> tuple[Storage, str]:
    fs, path_prefix = _get_bucket_fs(bucket_name, papai_fs, list_buckets)
    if path_prefix != "":
        path_prefix = joinpath(bucket_name, path_prefix)

    return fs, path_prefix


def list_bucket_objects(
    bucket_name: str,
    list_buckets: list[Bucket],
    papai_fs: Storage,
    path: str = "",
    recursive: bool = True,
):
    """List bucket objects with name starting with `path`.

    Parameters
    ----------
    bucket_name : str
        Name of the bucket to list in.
    list_buckets : list[Bucket]
        List of all bucket configurations.
    papai_fs : Storage
        Default filesystem to interact with the bucket.
    path : str, optional
        List all objects that starts with `path`, by default "".
    recursive : bool, optional
        Whether to list objects that are deeper than `path`, by default True.
    """
    fs, bucket_name = _get_bucket_fs_with_prefix(bucket_name, papai_fs, list_buckets)
    full_path = joinpath(bucket_name, path)
    return fs.list_files(full_path, recursive)


def glob_bucket_objects(
    bucket_name: str,
    list_buckets: list[Bucket],
    papai_fs: Storage,
    pattern: str = "",
    recursive: bool = True,
):
    """List files in a remote directory that match a pattern.

    Parameters
    ----------
    bucket_name : str
        Name of the bucket to list in.
    list_buckets : list[Bucket]
        List of all bucket configurations.
    papai_fs : Storage
        Default filesystem to interact with the bucket.
    prefix : str, optional
        List all objects that starts with `prefix`, by default "".
    recursive : bool, optional
        Whether to list objects that are deeper than `prefix`, by default True.
    """
    fs, bucket_name = _get_bucket_fs_with_prefix(bucket_name, papai_fs, list_buckets)
    full_path = joinpath(bucket_name, pattern)
    return fs.glob_files(full_path, recursive)


def read_from_bucket_to_file(
    bucket_name: str,
    object_name: str,
    list_buckets: list[Bucket],
    papai_fs: Storage,
    destination_directory: str | None = None,
) -> str:
    warnings.warn(
        "read_from_bucket_to_file is deprecated, use get_from_bucket instead",
        DeprecationWarning,
    )
    return get_from_bucket(
        bucket_name,
        object_name,
        list_buckets,
        papai_fs,
        destination_directory=destination_directory,
    )


def get_from_bucket(
    bucket_name: str,
    object_names: str | list[str],
    list_buckets: list[Bucket],
    papai_fs: Storage,
    destination_directory: str | None = None,
) -> str:
    """Download files or folders from a bucket to the local file system.

    Parameters
    ----------
    bucket_name : str
        Name of the bucket to download from.
    object_names : str
        Path to the objects to download. It can be a glob pattern or a list of
        patterns.
    list_buckets : list[Bucket]
        List of all bucket configurations.
    papai_fs : Storage
        Default filesystem to interact with the bucket.
    destination_directory : str|None
        Directory in which to save the downloaded file. By default None.

    Returns
    -------
    str
        Local path to the directory containing downloaded objects.
    """
    fs, bucket_name = _get_bucket_fs_with_prefix(bucket_name, papai_fs, list_buckets)
    paths = joinpath(bucket_name, object_names)

    tmp_dir = generate_temporary_filename(destination_directory)
    fs.get(paths, tmp_dir)

    return tmp_dir


def put_to_bucket(
    bucket_name: str,
    object_names: str | list[str],
    list_buckets: list[Bucket],
    papai_fs: Storage,
    file_path: str | bytes | os.PathLike,
):
    """Upload files / folders to a bucket.

    Parameters
    ----------
    bucket_name : str
        Name of the bucket to upload to.
    object_names : str
        Path to the objects to download. It can be a glob pattern or a list of
        patterns.
    list_buckets : list[Bucket]
        List of all bucket configurations.
    papai_fs : Storage
        Default filesystem to interact with the bucket.
    file_path : str | bytes | os.PathLike
        File to upload to the bucket.
    """
    fs, bucket_name = _get_bucket_fs_with_prefix(bucket_name, papai_fs, list_buckets)
    paths = joinpath(bucket_name, object_names)

    fs.put(file_path, paths)


def write_to_file_in_bucket(
    bucket_name: str,
    file_name: str,
    list_buckets: list[Bucket],
    papai_fs: Storage,
    data: io.BytesIO | str,
):
    """Upload a file to a bucket / Write data in a bucket.

    Parameters
    ----------
    bucket_name : str
        Name of the bucket to write in.
    file_name : str
        Path to which the data will be written.
    list_buckets : list[Bucket]
        List of all bucket configurations.
    papai_fs : Storage
        Default filesystem to interact with the bucket.
    data : io.BytesIO | str
        Data to write to the file
    """
    fs, bucket_name = _get_bucket_fs_with_prefix(bucket_name, papai_fs, list_buckets)
    path = joinpath(bucket_name, file_name)

    fs.write_to_file(path, content=data)


def write_file_in_bucket(
    bucket_name: str,
    file_name: str,
    list_buckets: list[Bucket],
    papai_fs: Storage,
    data: io.BytesIO | str | None = None,
    file_path: str | bytes | os.PathLike | None = None,
):
    if file_path is not None:
        warnings.warn(
            "write_file_in_bucket with `file_path` argument is deprecated, use put_to_bucket instead",
            DeprecationWarning,
        )
        put_to_bucket(bucket_name, file_name, list_buckets, papai_fs, file_path)

    if data is not None:
        warnings.warn(
            "write_file_in_bucket with `data` argument is deprecated, use write_to_file_in_bucket instead",
            DeprecationWarning,
        )
        write_to_file_in_bucket(bucket_name, file_name, list_buckets, papai_fs, data)

    raise ValueError("You must provide either data or a file_path")


def import_dataset(
    dataset_name: str, list_parquets: list[Parquet], papai_fs: Storage
) -> pd.DataFrame:
    """Load a pandas DataFrame from a parquet file in the papai bucket.

    Parameters
    ----------
    dataset_name : str
        Name of the dataset to load.
    list_parquets : list[Parquet]
        List of all parquet configurations.
    papai_fs : Storage
        Papai filesystem in which to find the dataset.

    Returns
    -------
    pd.DataFrame
    """
    parquet = _get_io_config("step_name", dataset_name, list_parquets, "dataset")
    parquet_path = joinpath(parquet["bucket_name"], parquet["object_name"])
    return papai_fs.read_dataset_from_parquet(parquet_path)


def export_dataset(
    dataset: pd.DataFrame,
    dataset_name: str,
    list_parquets: list[Parquet],
    papai_fs: Storage,
):
    """Write a pandas DataFrame to a parquet file in the papai bucket.

    Parameters
    ----------
    dataset : pd.DataFrame
        Pandas DataFrame to write to the bucket.
    dataset_name : str
        Name of the dataset to load.
    list_parquets : list[Parquet]
        List of all parquet configurations.
    papai_fs : Storage
        Papai filesystem in which to find the dataset.
    """
    parquet = _get_io_config("step_name", dataset_name, list_parquets, "dataset")
    parquet_path = joinpath(parquet["bucket_name"], parquet["object_name"])
    papai_fs.write_dataframe_to_parquet(parquet_path, dataset)


def _get_artefacts_folder_path(
    registry_name: str,
    list_registries: list[Registry],
    run_uuid: str | None = None,
) -> str:
    """
    Get the path to the artefacts folder in the papai filesystem. If a run_uuid
    is provided, the default run_uuid will be replaced by the provided one.
    """
    registry = _get_io_config(
        "registry_name", registry_name, list_registries, "registry"
    )
    artefacts_folder_path = joinpath(
        registry["bucket_name"], registry["artefacts_path"]
    )

    if run_uuid is not None:
        # Replace the default run_uuid by the provided one
        artefacts_folder_path = artefacts_folder_path.rsplit("/", 1)[0] + "/" + run_uuid

    return artefacts_folder_path


def get_model_artefacts(
    registry_name: str,
    artefact_path: str,
    run_uuid: str,
    registry_inputs: list[Registry],
    papai_fs: Storage,
    destination_directory: str | None = None,
) -> str:
    """Download a model artefact from the papai filesystem.

    Parameters
    ----------
    registry_name : str
        Name of the registry to get the artefact from.
    artefact_path : str
        Path to the artefact.
    registry_inputs : list[Registry]
        List of all registry configurations.
    papai_fs : Storage
        PapAI filesystem to interact with the bucket.
    run_uuid : str | None
        The UUID of the run you want to get artefacts from. If None, it will
        be automatically be the activated run. By default None.
    destination_directory : str | None
        Directory in which to save the downloaded file. By default None.

    Returns
    -------
    str
        Local path to the downloaded artefact.
    """
    artefacts_folder_path = _get_artefacts_folder_path(
        registry_name, registry_inputs, run_uuid
    )
    artefact_full_path = joinpath(artefacts_folder_path, artefact_path)

    tmp_file = generate_temporary_filename(destination_directory)
    papai_fs.get_file(artefact_full_path, tmp_file)

    return tmp_file


def save_model_artefact(
    data,
    registry_name: str,
    artefact_path: str,
    registry_inputs: list[Registry],
    papai_fs: Storage,
    run_uuid: str | None = None,
):
    """Upload an artefact to the papai filesystem.

    Parameters
    ----------
    data: str | bytes
        Data to write to the file.
    registry_name : str
        Name of the registry to get the artefact from.
    artefact_path : str
        Path to the artefact.
    registry_inputs : list[Registry]
        List of all registry configurations.
    papai_fs : Storage
        PapAI filesystem to interact with the bucket.
    run_uuid : str | None
        The UUID of the run you want to save artefacts to. If None, it will
        be automatically be the activated run. By default None.

    Returns
    -------
    str
        Local path to the downloaded artefact.
    """
    artefacts_folder_path = _get_artefacts_folder_path(
        registry_name, registry_inputs, run_uuid
    )
    artefact_full_path = joinpath(artefacts_folder_path, artefact_path)

    papai_fs.write_to_file(artefact_full_path, data)
